// EigenLF.c : Defines the entry point for the console application.

//				Tests the Eigen function

#include <stdlib.h>
#include <stdbool.h>	// for bool
#include "Algo358.h"
#include <time.h>

void Eigen(float _Complex a[][maxA], int n, float _Complex va[], float _Complex ve[][maxA], bool herm, bool *error);
void CSVD(float _Complex a[][maxA], int m, int n, int p, int nu, int nv,
		float s[], float _Complex u[][maxA], float _Complex v[][maxA]);
void TestEigen();
void TimeStamp();
void WriteMatrix(float _Complex [][maxA], int, int);
void WriteMatrixZ(float _Complex [][maxA], int, int);
void MatMulSq(int n, float _Complex a[][maxA], float _Complex b[][maxA], float _Complex p[][maxA]);
void Hermitian(float _Complex a[][maxA], int n);

int ti;			// test case number, 1 to 4
double time0;

// ------- initialize
void init(void)
{
	char name[]    = "EigenLF - Computes Eigenvalues and Eigenvectors of a square complex matrix";
	char version[] = "Lcc-Win32 Console Version 1.1.1";
	char author1[] = "Copyright (c) 1981-2007 by author: Harry J. Smith,";
	char author2[] = "19628 Via Monte Dr., Saratoga, CA 95070.  All rights reserved.";
	char usage1[]  = "Usage: EigenLF [1 to 4] [>Test.out]";
	printf("\n%s\n", name);		// show credits
	printf("%s\n", version);
	printf("%s\n", author1);
	printf("%s\n\n", author2);
	printf("%s\n\n", usage1);	// show usage
} // init

// The main entry point for the application EigenLF
//int _tmain(int argc, _TCHAR* argv[])
int main(int argc, char* argv[])
{
	init();
	char tc[5] = "1";
	if (argc > 1)							// if arguments
		strncpy(tc, argv[1], 2);		// Copy string from user input line
	ti = atoi(tc);
	if (ti < 2 || ti > 4)
		ti = 1;
	printf("Running test case %i of 4\n\n", ti);
	time0 = (double)clock() / (double)CLOCKS_PER_SEC;
	TimeStamp();
	printf("\nEigenLF:\n");
	TestEigen();
	printf("EigenLF:\n");
	printf("Normal end of execution\n\n");
	TimeStamp();
	return 0;
} // main

float _Complex SetTo(float x, float y)		// Set Real and Imag parts
{
	return x + I * y;
} // SetTo

void TestEigen()		//test Eigen
{
	int i, j, n;
	bool herm;
	bool error;
	float _Complex a[maxA][maxA];
	float _Complex u[maxA][maxA];
	float _Complex v[maxA][maxA];
	float _Complex w[maxA][maxA];
	float _Complex va[maxA];
	float _Complex ve[maxA][maxA];
	float _Complex cZero = SetTo(0.0F, 0.0F);
	a[0][0] = SetTo( 0.4499F, -0.1267F);
	a[0][1] = SetTo(-0.2361F,  0.0775F);
	a[0][2] = SetTo( 0.5008F, -0.7799F);
	a[0][3] = SetTo(-0.7702F, -0.3143F);
	a[0][4] = SetTo(-0.5084F,  0.5621F);
	a[0][5] = SetTo(-0.5577F, -0.0913F);
	a[0][6] = SetTo(-0.7207F, -0.9798F);
	a[0][7] = SetTo(-0.5367F,  0.4677F);
	a[1][0] = SetTo(-0.8432F, -0.3443F);
	a[1][1] = SetTo( 0.0186F, -0.6332F);
	a[1][2] = SetTo( 0.3505F,  0.0166F);
	a[1][3] = SetTo(-0.8892F,  0.2657F);
	a[1][4] = SetTo(-0.5066F,  0.6005F);
	a[1][5] = SetTo( 0.1421F,  0.0154F);
	a[1][6] = SetTo(-0.2327F, -0.2606F);
	a[1][7] = SetTo(-0.4535F,  0.4458F);
	a[2][0] = SetTo( 0.5896F,  0.2601F);
	a[2][1] = SetTo( 0.8928F,  0.0103F);
	a[2][2] = SetTo( 0.4350F, -0.2666F);
	a[2][3] = SetTo(-0.7799F, -0.5512F);
	a[2][4] = SetTo(-0.1022F, -0.4500F);
	a[2][5] = SetTo(-0.6975F, -0.8781F);
	a[2][6] = SetTo( 0.3364F, -0.8415F);
	a[2][7] = SetTo( 0.3890F,  0.6854F);
	a[3][0] = SetTo( 0.3911F,  0.3234F);
	a[3][1] = SetTo(-0.5605F,  0.7638F);
	a[3][2] = SetTo(-0.2009F,  0.2707F);
	a[3][3] = SetTo( 0.0314F, -0.4336F);
	a[3][4] = SetTo(-0.1046F,  0.3267F);
	a[3][5] = SetTo(-0.8820F, -0.9251F);
	a[3][6] = SetTo( 0.4480F, -0.0054F);
	a[3][7] = SetTo( 0.5823F, -0.2339F);
	a[4][0] = SetTo(-0.1395F, -0.1561F);
	a[4][1] = SetTo( 0.3064F,  0.0263F);
	a[4][2] = SetTo(-0.0975F,  0.9019F);
	a[4][3] = SetTo( 0.2491F,  0.5787F);
	a[4][4] = SetTo( 0.4106F, -0.8102F);
	a[4][5] = SetTo(-0.7131F,  0.5675F);
	a[4][6] = SetTo( 0.0759F,  0.1507F);
	a[4][7] = SetTo( 0.1801F, -0.3816F);
	a[5][0] = SetTo(-0.5889F, -0.0238F);
	a[5][1] = SetTo( 0.9861F, -0.4593F);
	a[5][2] = SetTo(-0.5219F,  0.8539F);
	a[5][3] = SetTo( 0.6382F, -0.0232F);
	a[5][4] = SetTo( 0.8741F,  0.6109F);
	a[5][5] = SetTo(-0.2208F, -0.1206F);
	a[5][6] = SetTo(-0.0868F,  0.7716F);
	a[5][7] = SetTo(-0.4976F,  0.4122F);
	a[6][0] = SetTo(-0.3180F,  0.8621F);
	a[6][1] = SetTo( 0.3905F, -0.8872F);
	a[6][2] = SetTo( 0.8140F, -0.2542F);
	a[6][3] = SetTo(-0.2290F,  0.4111F);
	a[6][4] = SetTo(-0.9723F, -0.4355F);
	a[6][5] = SetTo( 0.5195F,  0.4954F);
	a[6][6] = SetTo(-0.8843F,  0.4552F);
	a[6][7] = SetTo( 0.9156F, -0.5878F);
	a[7][0] = SetTo( 0.1596F, -0.2962F);
	a[7][1] = SetTo( 0.0272F,  0.3327F);
	a[7][2] = SetTo(-0.5806F, -0.1952F);
	a[7][3] = SetTo(-0.9280F, -0.5908F);
	a[7][4] = SetTo( 0.6263F,  0.0384F);
	a[7][5] = SetTo( 0.0210F, -0.7325F);
	a[7][6] = SetTo(-0.9907F, -0.9717F);
	a[7][7] = SetTo( 0.0878F, -0.9056F);
	if (ti == 1)
	{
		n = 5;
		herm = false;
	}
	if (ti == 2)
	{
		n = 5;
		herm = true;
		Hermitian(a, n);	// Set the matrix
	}
	if (ti == 3)
	{
		n = 8;
		herm = false;
	}
	if (ti == 4)
	{
		n = 8;
		Hermitian(a, n);	// Set the matrix
		herm = true;
	}
		printf("Call the Eigen function to compute the Eigenvalues and Eigenvectors of\n");
	if (herm)
		printf("an N by N Hermitian");
	else
		printf("an N by N complex");
	printf(" matrix A, N = %i.\n", n);
	printf("\nMatrix A:\n\n");
	WriteMatrix(a, n, n);
	int p = 0;
	int nu = n;
	int nv = n;

	// INITIAL TESTS
	void PolNorm(float _Complex a[], int *n);
	float _Complex pp[9];
	pp[0] = SetTo( 0.0000F,  0.0000F);
	pp[1] = SetTo(-0.2009F,  0.2707F);
	pp[2] = SetTo( 0.4106F, -0.8102F);
	int ppn = 3;
	PolNorm(pp, &ppn);
	void Laguer(float _Complex a[], int n, int *its, float _Complex *x);
	pp[0] = SetTo( 0.4499F, -0.1267F);
	pp[1] = SetTo(-0.2009F,  0.2707F);
	pp[2] = SetTo( 0.4106F, -0.8102F);
	int its;
	// answer 0.80826026, 1.0269444
	float _Complex root;
	bool polish = true;
	Laguer(pp, 3, &its, &root);
	void PolRoots(float _Complex p[], int n, bool polish, float _Complex r[]);
	float _Complex roots[2];
	PolRoots(pp, 3, polish, roots);
	void CharPoly(float _Complex a[][maxA], int n, float _Complex p[]);
	CharPoly(a, n, pp);
	// END OF INITIAL TESTS

	Eigen(a, n, va, ve, herm, &error);
	printf("Eigenvalues va:\n\n");
	for (i = 0; i < n; i++)
		printf("  (%.7f,%.7f)\n", crealf(va[i]), cimagf(va[i]));
	printf("\nEigenvectors ve:\n\n");
	WriteMatrix(ve, n, n);
	MatMulSq(n, a, ve, u);
	printf("A * ve:\n\n");
	WriteMatrix(u, n, n);
	for (i = 0; i < n; i++)
		for (j = 0; j < n; j++)
			w[i][j] = cZero;
	for (i = 0; i < n; i++)
	{
		float _Complex tw = va[i];		// Lcc-Win32 had a problem with w[i][i] = va[i];
		w[i][i] = tw;
	}
	MatMulSq(n, ve, w, v);
	float _Complex maxErr = cZero;
	for (i = 0; i < n; i++)
		for (j = 0; j < n; j++)
		{
			w[i][j] = u[i][j] - v[i][j];
			if (fabs(crealf(w[i][j])) > crealf(maxErr))
				maxErr = SetTo(fabsf(crealf(w[i][j])), cimagf(maxErr));
			if (fabs(cimagf(w[i][j])) > cimagf(maxErr))
				maxErr = SetTo(crealf(maxErr), fabsf(cimagf(w[i][j])));
		}
	printf("A * ve - ve * Diag(va), should be zero: Max error = (%.3e,%.3e)\n\n", crealf(maxErr), cimagf(maxErr));
	WriteMatrixZ(w, n, n);
} // TestEigen

void Hermitian(float _Complex a[][maxA], int n)
// make a an n by n Hermitian matrix. use upper triangle to make lower triangle
{
	int i, j, k;
	for (i = 0; i < n - 1; i++)
	{
		for (j = i + 1; j < n; j++)
			a[j][i] = conjf(a[i][j]);
	}
	for (k = 0; k < n; k++)
		a[k][k] = SetTo(crealf(a[k][k]), 0.0F);

} // Hermitian

void MatMulSq(int n, float _Complex a[][maxA], float _Complex b[][maxA], float _Complex p[][maxA])
// p = a * b. a, b, p all n by n complex matrices
{
	int i, j, k;
	float _Complex cZero = SetTo(0.0F, 0.0F);
	for (i = 0; i < n; i++)
		for (j = 0; j < n; j++)
		{
			p[i][j] = cZero;
			for (k = 0; k < n; k++)
			{
				p[i][j] = p[i][j] + a[i][k] * b[k][j];
			}
		}
} // MatMulSq

void WriteMatrix(float _Complex a[][maxA], int m, int n)
// print an m by n complex matrix
{
	int i, j;
	for (i = 0; i < m; i++)
	{
		for (j = 0; j < n; j++)
		{
			printf(" ");
			if (crealf(a[i][j]) >= 0.0F)
				printf(" ");
			else
				printf("-");
			if (crealf(a[i][j]) == 0.0F)
				printf("0.0000");
			else
				printf("%.4f", fabsf(crealf(a[i][j])));
			if (cimagf(a[i][j]) >= 0.0F)
				printf("+");
			else
				printf("-");
			if (cimagf(a[i][j]) == 0.0F)
				printf("0.0000");
			else
				printf("%.4f", fabsf(cimagf(a[i][j])));
		}
		printf("\n");
	}
	printf("\n");
} // WriteMatrix

void WriteMatrixZ(float _Complex a[][maxA], int m, int n)
// print an m by n complex matrix that is near zero
{
	int i, j;
	for (i = 0; i < m; i++)
	{
		for (j = 0; j < n; j++)
		{
			printf(" ");
			if (crealf(a[i][j]) >= 0.0F)
				printf(" ");
			else
				printf("-");
			if (crealf(a[i][j]) == 0.0F)
				printf("0.0000");
			else
				printf("%.0e", fabsf(crealf(a[i][j])));
			if (cimagf(a[i][j]) >= 0.0F)
				printf("+");
			else
				printf("-");
			if (cimagf(a[i][j]) == 0.0F)
				printf("0.0000");
			else
				printf("%.0e", fabsf(cimagf(a[i][j])));
		}
		printf("\n");
	}
	printf("\n");
} // WriteMatrixZ

void TimeStamp()
// printout the current MM/dd/yyyy hh:mm:ss AM/PM  s.ss sec. as a timestamp
{
	int i;
	double time1 = (double)clock() / (double)CLOCKS_PER_SEC;
	double timeD = time1 - time0;
	time0 = time1;
	char date[9];
	struct tm *newtime;
	time_t ltime;
	/* get time in seconds */
	time(&ltime);
	/* convert it to the structure tm */
	newtime = localtime(&ltime);
	char sta[30];
	for (i = 0; i < 24; i++)
	{
		sta[i] = (char)asctime(newtime)[i];
	}
	sta[24] = 0;
	printf("%s  %.2f sec.\n", sta, timeD);
} // TimeStamp

/* end of file EigenLF.c ------------------------------------------ */
